@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # This file is part of urlwatch (https://thp.io/2008/urlwatch/).
-# Copyright (c) 2008-2018 Thomas Perl <m@thp.io>
+# Copyright (c) 2008-2019 Thomas Perl <m@thp.io>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -134,6 +134,20 @@ class JobBase(object, metaclass=TrackSubClasses):
     def __repr__(self):
         return '<%s %s>' % (self.__kind__, ' '.join('%s=%r' % (k, v) for k, v in list(self.to_dict().items())))
 
+    def _set_defaults(self, defaults):
+        if isinstance(defaults, dict):
+            for key, value in defaults.items():
+                if key in self.__optional__ and getattr(self, key) is None:
+                    setattr(self, key, value)
+
+    def with_defaults(self, config):
+        new_job = JobBase.unserialize(self.serialize())
+        cfg = config.get('job_defaults')
+        if isinstance(cfg, dict):
+            new_job._set_defaults(cfg.get(self.__kind__))
+            new_job._set_defaults(cfg.get('all'))
+        return new_job
+
     def get_guid(self):
         location = self.get_location()
         sha_hash = hashlib.new('sha1')
@@ -143,10 +157,19 @@ class JobBase(object, metaclass=TrackSubClasses):
     def retrieve(self, job_state):
         raise NotImplementedError()
 
+    def format_error(self, exception, tb):
+        return tb
+
+    def ignore_error(self, exception):
+        return False
+
 
 class Job(JobBase):
     __required__ = ()
-    __optional__ = ('name', 'filter', 'max_tries', 'diff_tool')
+    __optional__ = ('name', 'filter', 'max_tries', 'diff_tool', 'compared_versions')
+
+    # determine if hyperlink "a" tag is used in HtmlReporter
+    LOCATION_IS_URL = False
 
     def pretty_name(self):
         return self.name if self.name else self.get_location()
@@ -180,8 +203,9 @@ class UrlJob(Job):
 
     __required__ = ('url',)
     __optional__ = ('cookies', 'data', 'method', 'ssl_no_verify', 'ignore_cached', 'http_proxy', 'https_proxy',
-                    'headers', 'ignore_connection_errors')
+                    'headers', 'ignore_connection_errors', 'ignore_http_error_codes', 'encoding', 'timeout')
 
+    LOCATION_IS_URL = True
     CHARSET_RE = re.compile('text/(html|plain); charset=([^;]*)')
 
     def get_location(self):
@@ -203,7 +227,7 @@ class UrlJob(Job):
         if job_state.timestamp is not None:
             headers['If-Modified-Since'] = email.utils.formatdate(job_state.timestamp)
 
-        if self.ignore_cached:
+        if self.ignore_cached or job_state.tries > 0:
             headers['If-None-Match'] = None
             headers['If-Modified-Since'] = email.utils.formatdate(0)
             headers['Cache-Control'] = 'max-age=172800'
@@ -229,13 +253,23 @@ class UrlJob(Job):
         if self.headers:
             self.add_custom_headers(headers)
 
+        if self.timeout is None:
+            # default timeout
+            timeout = 60
+        elif self.timeout == 0:
+            # never timeout
+            timeout = None
+        else:
+            timeout = self.timeout
+
         response = requests.request(url=self.url,
                                     data=self.data,
                                     headers=headers,
                                     method=self.method,
                                     verify=(not self.ssl_no_verify),
                                     cookies=self.cookies,
-                                    proxies=proxies)
+                                    proxies=proxies,
+                                    timeout=timeout)
 
         response.raise_for_status()
         if response.status_code == requests.codes.not_modified:
@@ -249,7 +283,7 @@ class UrlJob(Job):
         # urlwatch behavior and try UTF-8 decoding first.
         content_type = response.headers.get('Content-type', '')
         content_type_match = self.CHARSET_RE.match(content_type)
-        if not content_type_match:
+        if not content_type_match and not self.encoding:
             try:
                 try:
                     try:
@@ -261,6 +295,8 @@ class UrlJob(Job):
             except LookupError:
                 # If this is an invalid encoding, decode as ascii (Debian bug 731931)
                 return response.content.decode('ascii', 'ignore')
+        if self.encoding:
+            response.encoding = self.encoding
 
         return response.text
 
@@ -274,6 +310,27 @@ class UrlJob(Job):
             headers.pop(header, None)
         headers.update(self.headers)
 
+    def format_error(self, exception, tb):
+        if isinstance(exception, requests.exceptions.RequestException):
+            # Instead of a full traceback, just show the HTTP error
+            return str(exception)
+        return tb
+
+    def ignore_error(self, exception):
+        if isinstance(exception, requests.exceptions.ConnectionError) and self.ignore_connection_errors:
+            return True
+        elif isinstance(exception, requests.exceptions.HTTPError):
+            status_code = exception.response.status_code
+            ignored_codes = []
+            if isinstance(self.ignore_http_error_codes, int) and self.ignore_http_error_codes == status_code:
+                return True
+            elif isinstance(self.ignore_http_error_codes, str):
+                ignored_codes = [s.strip().lower() for s in self.ignore_http_error_codes.split(',')]
+            elif isinstance(self.ignore_http_error_codes, list):
+                ignored_codes = [str(s).strip().lower() for s in self.ignore_http_error_codes]
+            return str(status_code) in ignored_codes or '%sxx' % (status_code // 100) in ignored_codes
+        return False
+
 
 class BrowserJob(Job):
     """Retrieve an URL, emulating a real web browser"""
@@ -282,6 +339,8 @@ class BrowserJob(Job):
 
     __required__ = ('navigate',)
 
+    LOCATION_IS_URL = True
+
     def get_location(self):
         return self.navigate
 
