@@ -1,6 +1,6 @@
 #
 # This file is part of urlwatch (https://thp.io/2008/urlwatch/).
-# Copyright (c) 2008-2018 Thomas Perl <m@thp.io>
+# Copyright (c) 2008-2019 Thomas Perl <m@thp.io>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -35,9 +35,11 @@ import shlex
 import email.utils
 import itertools
 import logging
+import os
 import sys
 import time
-import cgi
+import html
+import functools
 
 import requests
 
@@ -61,7 +63,7 @@ logger = logging.getLogger(__name__)
 
 # Regular expressions that match the added/removed markers of GNU wdiff output
 WDIFF_ADDED_RE = r'[{][+].*?[+][}]'
-WDIFF_REMOVED_RE = r'[[][-].*?[-][]]'
+WDIFF_REMOVED_RE = r'[\[][-].*?[-][]]'
 
 
 class ReporterBase(object, metaclass=TrackSubClasses):
@@ -108,19 +110,20 @@ class ReporterBase(object, metaclass=TrackSubClasses):
 
     def unified_diff(self, job_state):
         if job_state.job.diff_tool is not None:
-            with tempfile.NamedTemporaryFile() as old_file, tempfile.NamedTemporaryFile() as new_file:
-                old_file.write(job_state.old_data.encode('utf-8'))
-                old_file.flush()
-                new_file.write(job_state.new_data.encode('utf-8'))
-                new_file.flush()
-                cmdline = shlex.split(job_state.job.diff_tool) + [old_file.name, new_file.name]
+            with tempfile.TemporaryDirectory() as tmpdir:
+                old_file_path = os.path.join(tmpdir, 'old_file')
+                new_file_path = os.path.join(tmpdir, 'new_file')
+                with open(old_file_path, 'w+b') as old_file, open(new_file_path, 'w+b') as new_file:
+                    old_file.write(job_state.old_data.encode('utf-8'))
+                    new_file.write(job_state.new_data.encode('utf-8'))
+                cmdline = shlex.split(job_state.job.diff_tool) + [old_file_path, new_file_path]
                 proc = subprocess.Popen(cmdline, stdout=subprocess.PIPE)
                 stdout, _ = proc.communicate()
                 # Diff tools return 0 for "nothing changed" or 1 for "files differ", anything else is an error
                 if proc.returncode in (0, 1):
                     return stdout.decode('utf-8')
                 else:
-                    raise subprocess.CalledProcessError(result, cmdline)
+                    raise subprocess.CalledProcessError(proc.returncode, cmdline)
 
         timestamp_old = email.utils.formatdate(job_state.timestamp, localtime=1)
         timestamp_new = email.utils.formatdate(time.time(), localtime=1)
@@ -137,8 +140,8 @@ class SafeHtml(object):
         return self.s
 
     def format(self, *args, **kwargs):
-        return str(self).format(*(cgi.escape(str(arg)) for arg in args),
-                                **{k: cgi.escape(str(v)) for k, v in kwargs.items()})
+        return str(self).format(*(html.escape(str(arg)) for arg in args),
+                                **{k: html.escape(str(v)) for k, v in kwargs.items()})
 
 
 class HtmlReporter(ReporterBase):
@@ -169,7 +172,7 @@ class HtmlReporter(ReporterBase):
         for job_state in self.report.get_filtered_job_states(self.job_states):
             job = job_state.job
 
-            if job.__kind__ == 'url':
+            if job.LOCATION_IS_URL:
                 title = '<a href="{location}">{pretty_name}</a>'
             elif job.pretty_name() != job.get_location():
                 title = '<span title="{location}">{pretty_name}</span>'
@@ -240,6 +243,15 @@ class TextReporter(ReporterBase):
         show_details = cfg['details']
         show_footer = cfg['footer']
 
+        if cfg['minimal']:
+            for job_state in self.report.get_filtered_job_states(self.job_states):
+                pretty_name = job_state.job.pretty_name()
+                location = job_state.job.get_location()
+                if pretty_name != location:
+                    location = '%s (%s)' % (pretty_name, location)
+                yield ': '.join((job_state.verb.upper(), location))
+            return
+
         summary = []
         details = []
         for job_state in self.report.get_filtered_job_states(self.job_states):
@@ -248,12 +260,12 @@ class TextReporter(ReporterBase):
             details.extend(details_part)
 
         if summary:
-            sep = line_length * '='
-            yield from itertools.chain(
+            sep = (line_length * '=') or None
+            yield from (part for part in itertools.chain(
                 (sep,),
                 ('%02d. %s' % (idx + 1, line) for idx, line in enumerate(summary)),
                 (sep, ''),
-            )
+            ) if part is not None)
 
         if show_details:
             yield from details
@@ -291,11 +303,12 @@ class TextReporter(ReporterBase):
 
         summary_part.append(pretty_summary)
 
-        sep = line_length * '-'
+        sep = (line_length * '-') or None
         details_part.extend((sep, summary, sep))
         if content is not None:
             details_part.extend((content, sep))
-        details_part.extend(('', ''))
+        details_part.extend(('', '') if sep else ('',))
+        details_part = [part for part in details_part if part is not None]
 
         return summary_part, details_part
 
@@ -305,8 +318,12 @@ class StdoutReporter(TextReporter):
 
     __kind__ = 'stdout'
 
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self._has_color = sys.stdout.isatty() and self.config.get('color', False)
+
     def _incolor(self, color_id, s):
-        if sys.stdout.isatty() and self.config.get('color', False):
+        if self._has_color:
             return '\033[9%dm%s\033[0m' % (color_id, s)
         return s
 
@@ -322,11 +339,19 @@ class StdoutReporter(TextReporter):
     def _blue(self, s):
         return self._incolor(4, s)
 
+    def _get_print(self):
+        if sys.platform == 'win32' and self._has_color:
+            from colorama import AnsiToWin32
+            return functools.partial(print, file=AnsiToWin32(sys.stdout).stream)
+        return print
+
     def submit(self):
+        print = self._get_print()
+
         cfg = self.report.config['report']['text']
         line_length = cfg['line_length']
 
-        separators = (line_length * '=', line_length * '-', '-- ')
+        separators = (line_length * '=', line_length * '-', '-- ') if line_length else ()
         body = '\n'.join(super().submit())
 
         for line in body.splitlines():
@@ -371,7 +396,7 @@ class EMailReporter(TextReporter):
             logger.debug('Not sending e-mail (no changes)')
             return
         if self.config['method'] == "smtp":
-            smtp_user = self.config['smtp'].get('user', self.config['from'])
+            smtp_user = self.config['smtp'].get('user', None) or self.config['from']
             mailer = SMTPMailer(smtp_user, self.config['smtp']['host'], self.config['smtp']['port'],
                                 self.config['smtp']['starttls'], self.config['smtp']['keyring'])
         elif self.config['method'] == "sendmail":
@@ -551,3 +576,44 @@ class TelegramReporter(TextReporter):
 
     def chunkstring(self, string, length):
         return (string[0 + i:length + i] for i in range(0, len(string), length))
+
+
+class SlackReporter(TextReporter):
+    """Custom Slack reporter"""
+    MAX_LENGTH = 4096
+
+    __kind__ = 'slack'
+
+    def submit(self):
+        webhook_url = self.config['webhook_url']
+        text = '\n'.join(super().submit())
+
+        if not text:
+            logger.debug('Not calling slack API (no changes)')
+            return
+
+        result = None
+        for chunk in self.chunkstring(text, self.MAX_LENGTH):
+            res = self.submit_to_slack(webhook_url, chunk)
+            if res.status_code != requests.codes.ok or res is None:
+                result = res
+
+        return result
+
+    def submit_to_slack(self, webhook_url, text):
+        logger.debug("Sending slack request with text:{0}".format(text))
+        post_data = {"text": text}
+        result = requests.post(webhook_url, json=post_data)
+        try:
+            if result.status_code == requests.codes.ok:
+                logger.info("Slack response: ok")
+            else:
+                logger.error("Slack error: {0}".format(result.text))
+        except ValueError:
+            logger.error(
+                "Failed to parse slack response. HTTP status code: {0}, content: {1}".format(result.status_code,
+                                                                                             result.content))
+        return result
+
+    def chunkstring(self, string, length):
+        return (string[0 + i:length + i] for i in range(0, len(string), length))
